// A simple singly-linked list implementation.  Every node stores a pointer to
// its successor in the list.  The list itself is represented as a pointer to
// its head.  Putting a strong emphasis on functional programming, Scala
// actually provides support for singly-linked lists natively.  The reason I
// offer a separate singly-linked list implementation here is twofold: (1) It
// makes the implementation of the list structure explicit, a bonus when talking
// about implementation of algorithms.  (2) If we don't insist on the
// implementation being functional (which we don't in this course), we can
// support two additional operations in constant time with almost no overhead in
// the representation of the data structure.  In particular, by also storing a
// pointer to the tail of the list, we can support a "concat" operation in
// constant time, which concatenates a given list to the end of a list.  The
// simpler version of this is appending a new element to the end of the list.
// The latter is particularly useful for implementing a queue.

package algos.ds

import scala.collection.AbstractIterator
import scala.collection.generic.Growable
import scala.collection.mutable.Builder
import scala.collection.mutable.GrowingBuilder

class SinglyLinkedList[T] extends Seq[T] {

  private class Node (
    val value : T,
    var succ  : Option[Node]
  )

  def apply(i : Int) : T = {
    var node = _head.get
    for (j <- 0 until i) node = node.succ.get
    node.value
  }

  override def isEmpty : Boolean = _head.isEmpty
  def          length  : Int     = _length

  override def head : T = _head.get.value
  override def last : T = _tail.get.value

  def append(x : T) {
    val n = Some(new Node(x, None))
    _length += 1
    _tail match {
      case None    => _head  = n
      case Some(t) => t.succ = n
    }
    _tail = n
  }

  def prepend(x : T) {
    _head = Some(new Node(x, _head))
    _length += 1
    if (_tail.isEmpty) _tail = _head
  }

  def removeHead() : T = {
    val oldHead = _head
    _head = _head.get.succ
    _length -= 1
    if (_head.isEmpty) _tail = None
    oldHead.get.value
  }

  def +=(x : T) = {
    append(x)
    this
  }

  def clear() {
    _head   = None
    _tail   = None
    _length = 0
  }

  def iterator : Iterator[T] = new ListIterator

  private class ListIterator extends AbstractIterator[T] {

    def hasNext : Boolean = !node.isEmpty

    def next : T = {
      val x = node.get.value
      node = node.get.succ
      x
    }

    private var node : Option[Node] = _head

  }

  private var _head   : Option[Node] = None
  private var _tail   : Option[Node] = None
  private var _length : Int          = 0

}
