package algos.introduction

import algos.ds.Queue
import java.io._
import scala.io.Source

object StableMatching {

  def main(args : Array[String]) {
    try {
      if (args.length != 2) throw new ArgsError
      val instance = read(args(0))
      val solution = instance.solve()
      write(solution, args(1))
    }
    catch {
      case e : Throwable => {
        println("ERROR: " + e.getMessage())
        usage()
        sys.exit(-1)
      }
    }
  }

  private def usage() {
    println("USAGE: scala StableMatching.jar <input file> <output file>");
    println();
    println("The input file format must be:");
    println("- 2n lines, each storing a permutation of {1, 2, ..., n}");
    println("  The first n lines are the preference lists of the men.");
    println("  The last n lines are the preference lists of the men.");
  }

  private def read(fileName : String) : Instance = {
    def parsePrefList(line : String) : Array[Int] =
      line.trim().split("\\s+").map(num => num.toInt - 1)
    val f = Source.fromFile(fileName)
    try {
      val prefLists                = f.getLines.map(parsePrefList).toArray
      val (mPrefLists, wPrefLists) = prefLists.splitAt(prefLists.length / 2)
      val men                      = for ((l,i) <- mPrefLists.zipWithIndex) yield new Man(i,l)
      val women                    = for ((l,i) <- wPrefLists.zipWithIndex) yield new Woman(i,l)
      new Instance(men, women)
    }
    finally {
      f.close()
    }
  }

  private def write(solution : Solution, fileName : String) {
    val f = new PrintWriter(new BufferedWriter(new FileWriter(fileName)))
    try {
      for ((w,m) <- solution) f.format("%d %d%n", new Integer(w+1), new Integer(m+1))
    }
    finally {
      f.close()
    }
  }

}

private object `package` {
  type Solution = Array[(Int, Int)]
}

private class Instance (

  private val men   : Array[Man],
  private val women : Array[Woman]

) {

  def solve() : Solution = {
    for (man <- men) unmarried.enqueue(man)
    while (!unmarried.isEmpty) {
      val man   = unmarried.dequeue()
      val woman = women(man.nextCandidate())
      woman.hearProposal(man) match {
        case Some(man1) => unmarried.enqueue(man1)
        case _          =>
      }
    }
    women.map(_.partner.get.name).zipWithIndex
  }

  private val unmarried = new Queue[Man]

}

private class Man (

  val name     : Int,
  val prefList : Array[Int]

) {

  def nextCandidate() : Int = {
    val cur = prefList(nextIndex)
    nextIndex += 1
    cur
  }

  private var nextIndex = 0

}

private class Woman (

  val name : Int,
  prefList : Array[Int]

) {

  def hearProposal(man : Man) : Option[Man] =
    if (iPrefer(man)) {
      val oldPartner = partner
      _partner = Some(man)
      oldPartner
    }
    else {
      Some(man)
    }

  private def iPrefer(man : Man) : Boolean = partner match {
    case None       => true
    case Some(man1) => invPrefList(man.name) < invPrefList(man1.name)
  }

  def partner : Option[Man] = _partner

  private def invertPrefList(prefList : Array[Int]) : Array[Int] = {
    val a = new Array[Int](prefList.length)
    for ((m,i) <- prefList.zipWithIndex) a(m) = i
    a
  }

  private val invPrefList : Array[Int]  = invertPrefList(prefList)
  private var _partner    : Option[Man] = None

}

class ArgsError extends Throwable("Wrong number of arguments")
